/*
 * geography.c - general purpose geography functions
 *
 * Copyright (C) 1999 Gordon D. Carrie
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 * 
 * Please address questions and bug reports to gcarri@mail.txcyber.com
 *
 ********************************************
 *
 */

#include <math.h>
#include "geography.h"

CartPoint LatLonToCart(GeoPoint geopt)
{
  /* Convert a latitude-longitude to Cartesian coordinates */
  /* Earth has unit radius */

  CartPoint cpt;	/* Hold return value */
  double
    coslat = cos(geopt.ord * RADPERDEG),
    sinlat = sin(geopt.ord * RADPERDEG),
    coslon = cos(geopt.abs * RADPERDEG),
    sinlon = sin(geopt.abs * RADPERDEG);

  cpt.x = coslat * coslon;
  cpt.y = coslat * sinlon;
  cpt.z = sinlat;
  return cpt;
}

double DomainLon(double lon, double refLon)
{
  /* Return a longitude in domain refLon - 180 <= lon <= refLon + 180 */

  /* Represent longitudes as integer number of "degree units"
   * and put in range [-360*units/deg, 360*units/deg] */
  int
    iLon    = DEGUNITS(lon) % UNITSPERGC,
    iRefLon = DEGUNITS(refLon) % UNITSPERGC;

  /* Put iRefLon in range [-180*UNITSPERDEG, 180*UNITSPERDEG] */
  iRefLon = (iRefLon < -UNITSPERGC_2)
            ? iRefLon += UNITSPERGC : (iRefLon > UNITSPERGC_2)
				      ? iRefLon -= UNITSPERGC : iRefLon;
  /* Put iLon in range [iRefLon - 180*units/deg, iRefLon + 180*units/deg] */
  iLon = (iLon < iRefLon - UNITSPERGC_2)
         ? iLon += UNITSPERGC : (iLon > iRefLon + UNITSPERGC_2)
	                        ? iLon -= UNITSPERGC : iLon;
  return iLon / (double)UNITSPERDEG;
}

double GwchLon(double lon)
{
  /* Put lon in "conventional" Greenwich-centered domain -180 < lon < 180 */
  return DomainLon(lon, 0.0);
}

GeoPoint DomainLonPt(GeoPoint geoPoint, double refLon)
{
  /* Return geoPoint with longitude in domain [refLon - 180, refLon + 180] */
  geoPoint.abs = DomainLon(geoPoint.abs, refLon);
  return geoPoint;
}

GeoPoint GwchLonPt(GeoPoint geoPt)
{
  /* Return geoPt with -180.0 <= geoPt.abs <= 180.0 */
  return DomainLonPt(geoPt, 0.0);
}

enum LonSgn LonCmp(double lon0, double lon1)
{
  /* Return:
   *   West if lon0 is west of lon1,
   *   East if lon0 is east of lon1,
   *   PrMd if lon0 and lon1 are the same longitude */

  /* Represent longitudes as integer number of "degree units"
   * and put in range [-360*UNITSPERDEG, 360*UNITSPERDEG] */
  int
    iLon0 = DEGUNITS(lon0) % UNITSPERGC,
    iLon1 = DEGUNITS(lon1) % UNITSPERGC;

  /* Put iLon1 in range [-180*units/deg, 180*units/deg] */
  iLon1 = (iLon1 < -UNITSPERGC_2)
            ? iLon1 += UNITSPERGC : (iLon1 > UNITSPERGC_2)
				      ? iLon1 -= UNITSPERGC : iLon1;
  /* Put iLon0 in range [iRefLon - 180*units/deg, iRefLon + 180*units/deg] */
  /* This is like puting the "Dateline" at iRefLon, for accurate comparison */
  iLon0 = (iLon0 < iLon1 - UNITSPERGC_2)
            ? iLon0 += UNITSPERGC : (iLon0 > iLon1 + UNITSPERGC_2)
	                            ? iLon0 -= UNITSPERGC : iLon0;

  /* Compare iLon0 to iLon1 */
  return (iLon0 < iLon1) ? West : (iLon0 > iLon1) ? East : PrMd;
}

enum LatSgn LatCmp(double lat0, double lat1)
{
  /* Determine if lat0 is south, north, or the same as lat1 */
  /* 
   * Return:
   *   North if lat0 is north of lat1
   *   South if lat0 is south of lat1
   *   Eq if lat0 is same as lat1
   *
   * Comparison is made at the precision implied by UNITSPERDEG
   *
   */

  int iLat0 = DEGUNITS(lat0), iLat1 = DEGUNITS(lat1);
  return (iLat0 < iLat1) ? South : (iLat0 > iLat1) ? North : Eq;
}

int DegCmp(double d0, double d1)
{
  /* Return:
   *   -1 if d0 < d1
   *    0 if d0 == d1
   *    1 if d0 > d1
   *   to within the accuracy implied by UNITSPERDEG */
  int i0 = DEGUNITS(d0), i1 = DEGUNITS(d1);
  return (i0 < i1) ? -1 : (i0 > i1) ? 1 : 0;
}

int LonBtwn(double lon, double lon0, double lon1)
{
  /*
   * Return true if lon0 < lon < lon1, or lon1 < lon < lon0
   * In this context, < means west.
   * In other words, return true if lon is "between" lon0 and lon1
   * Correctly handles intervals that cross the International Dateline.
   * Comparison is made at the precision implied by UNITSPERDEG.
   */

   int ilon, ilon0, ilon1;

   lon0 = DomainLon(lon0, lon);
   lon1 = DomainLon(lon1, lon);
   ilon0 = DEGUNITS(lon0);
   ilon1 = DEGUNITS(lon1);
   ilon  = DEGUNITS(lon);
   return    ((ilon0 > ilon1) ? ilon0-ilon1 : ilon1-ilon0) < UNITSPERGC_2
          && ((ilon0 < ilon && ilon < ilon1) || (ilon1 < ilon && ilon < ilon0));
}

int LonBtwn1(double lon, double lon0, double lon1)
{
  /*
   * Return true if lon0 <= lon < lon1, or lon1 < lon <= lon0
   * Like LonBtwn but includes one endpoint.  Useful with sequences of segments.
   * In this context, < means west.
   * In other words, return true if lon is "between" lon0 and lon1
   * Correctly handles intervals that cross the International Dateline.
   * Comparison is made at the precision implied by UNITSPERDEG.
   */

   int ilon, ilon0, ilon1;

   lon0 = DomainLon(lon0, lon);
   lon1 = DomainLon(lon1, lon);
   ilon0 = DEGUNITS(lon0);
   ilon1 = DEGUNITS(lon1);
   ilon  = DEGUNITS(lon);
   return    ((ilon0 > ilon1) ? ilon0-ilon1 : ilon1-ilon0) < UNITSPERGC_2
          && ((ilon0 < ilon && ilon <= ilon1)
	      || (ilon1 < ilon && ilon <= ilon0));
}

double
GeoDistance(GeoPoint p1, GeoPoint p2)
{
  /* GeoDistance - return distance between points p1 and p2	*/
  /*   in Great Circle degrees.					*/

  double cosDlon = cos((p2.abs - p1.abs) * RADPERDEG);
	
  return DEGPERRAD *
    acos(0.5 * ((cosDlon - 1) * cos((p2.ord + p1.ord) * RADPERDEG)
		+ (cosDlon + 1) * cos((p2.ord - p1.ord) * RADPERDEG)));
}

double
GeoQuickDistance(GeoPoint p1, GeoPoint p2)
{
  /* GeoQuickDistance - return distance between points p1 and p2	*/
  /* measured as absolute distance between the two points in space.	*/
  /* Similar to GeoDistance but faster.  Useful for comparisons.	*/

  double cosDlon = cos((p2.abs - p1.abs) * RADPERDEG);
	
  return 1.0 - (cosDlon - 1) * cos((p2.ord + p1.ord) * RADPERDEG)
	   - (cosDlon + 1) * cos((p2.ord - p1.ord) * RADPERDEG);
}

GeoPoint GeoStep(GeoPoint pnt, double dir, double dist)
{
  /*
   * Go distance dist (in great circle degrees) in direction dir (degrees)
   * from GeoPoint pnt (a lat-lon, in degrees).
   */
  
  double lon = pnt.abs * RADPERDEG;
  double lat = pnt.ord * RADPERDEG;
  GeoPoint endPt;				/* Hold return value */
  double cos_dist, sin_dist, cos_dir, sin_dir,
    cos_lat, cos_lon, sin_lon, sin_lat,
    x, y, z, h_1, h_2, dh;			/* Consolidating variables */
  
  dist *= RADPERDEG;
  dir *= RADPERDEG;
  cos_dist = cos(dist);
  sin_dist = sin(dist);
  cos_dir = cos(dir);
  sin_dir = sin(dir);
  cos_lat = cos(lat);
  cos_lon = cos(lon);
  sin_lon = sin(lon);
  sin_lat = sin(lat);
  x = cos_dist * cos_lon * cos_lat - sin_dir * sin_dist * sin_lon 
      - cos_lon * cos_dir * sin_dist * sin_lat;
  y = sin_dir * cos_lon * sin_dist + cos_dist * cos_lat * sin_lon 
      - cos_dir * sin_dist * sin_lon * sin_lat;
  z = cos_lat * cos_dir * sin_dist + cos_dist * sin_lat;
  
  /*  x^2  +  y^2  */
  h_1 = cos_dist * cos_lat - cos_dir * sin_dist * sin_lat;
  h_2 = (sin_dir) * (sin_dist);
  dh = h_1 * h_1 + h_2 * h_2;
  
  endPt.ord = (dh == 0.0) ? (z > 0 ? M_PI_2 : -M_PI_2)
                          : atan(z / sqrt(dh)) * DEGPERRAD;
  endPt.abs = atan2(y, x) * DEGPERRAD;
  
  return endPt;
}

/************ Computations with line segments ************/

GeoPoint GCircleX(GeoPoint ln1pt1, GeoPoint ln1pt2, GeoPoint ln2pt1,
  GeoPoint ln2pt2)
{
  /* Find intersections of two great circles, given two points on each.	*/
  /* Returns the intersection closer to the mean of the four points	*/
  /* (recall great circles have two intersections).			*/
  /* Returns bogus value if the two great circles are the same.		*/

  GeoPoint
    geoPoint,		/* Hold return value */
    nowhere = NOWHERE;	/* Error return value */

  /* Points on the great circles in Cartesian Geocentric coordinates */
  CartPoint ln1pt1c, ln1pt2c, ln2pt1c, ln2pt2c; 

  CartPoint
    p1,			/* Normal to plane containing ln1pt1 ln1pt2 */
    p2;			/* Normal to plane containing ln2pt2 ln2pt2 */

  CartPoint c;		/* Vector perpendicular to p1 and p2 */
  double cs;		/* Normalizing factor */

  CartPoint m;		/* Mean of the four input points */

  double dp, dm;	/* Distance from mean to +-(cx, cy, cz) */

  ln1pt1c = LatLonToCart(ln1pt1);
  ln1pt2c = LatLonToCart(ln1pt2);
  ln2pt1c = LatLonToCart(ln2pt1);
  ln2pt2c = LatLonToCart(ln2pt2);

  /* Normal to plane containing ln1pt1 and ln1pt2 (given by cross product) */
  p1.x = ln1pt1c.y * ln1pt2c.z - ln1pt1c.z * ln1pt2c.y;
  p1.y = ln1pt1c.z * ln1pt2c.x - ln1pt1c.x * ln1pt2c.z;
  p1.z = ln1pt1c.x * ln1pt2c.y - ln1pt1c.y * ln1pt2c.x;

  /* Normal to plane containing ln2pt1 and ln2pt2 (given by cross product) */
  p2.x = ln2pt1c.y * ln2pt2c.z - ln2pt1c.z * ln2pt2c.y;
  p2.y = ln2pt1c.z * ln2pt2c.x - ln2pt1c.x * ln2pt2c.z;
  p2.z = ln2pt1c.x * ln2pt2c.y - ln2pt1c.y * ln2pt2c.x;

  /* Perpendicular to p1 and p2 (another cross product) */
  c.x = p1.y * p2.z - p1.z * p2.y;
  c.y = p1.z * p2.x - p1.x * p2.z;
  c.z = p1.x * p2.y - p1.y * p2.x;

  /* If all pts on one line, undefined intersection */
  if (c.x == 0.0 && c.y == 0.0 && c.z == 0.0) {
    return nowhere;
  }

  /* Normalize */
  cs = 1.0 / sqrt(c.x * c.x + c.y * c.y + c.z * c.z);
  c.x *= cs;
  c.y *= cs;
  c.z *= cs;

  /* Intersection is at a pole */
  if (c.x == 0.0 && c.y == 0.0) {
    geoPoint.ord = (c.z > 0) ? 90.0 : -90.0;
    geoPoint.abs = 0.0;				/* Arbitrary lon at pole */
    return geoPoint;
  }

  /* Find which of the two intersections of the two great circles is	*/
  /* closer to the mean of the four input points			*/
  m.x = 0.25 * (ln1pt1c.x + ln1pt2c.x + ln2pt2c.x + ln2pt2c.x);
  m.y = 0.25 * (ln1pt1c.y + ln1pt2c.y + ln2pt2c.y + ln2pt2c.y);
  m.z = 0.25 * (ln1pt1c.z + ln1pt2c.z + ln2pt2c.z + ln2pt2c.z);
  dp = (m.x-c.x) * (m.x-c.x) + (m.y-c.y) * (m.y-c.y) + (m.z-c.z) * (m.z-c.z);
  dm = (m.x+c.x) * (m.x+c.x) + (m.y+c.y) * (m.y+c.y) + (m.z+c.z) * (m.z+c.z);
  if (dm < dp) {
    /* Mean point is closer to (-c.x, -c.y, -c.z) */
    c.x = -c.x;
    c.y = -c.y;
    c.z = -c.z;
  }

  geoPoint.ord = atan(c.z / sqrt(c.x * c.x + c.y * c.y)) * DEGPERRAD;
  geoPoint.abs = atan2(c.y, c.x) * DEGPERRAD;
  return geoPoint;
}

int GCircleXLatCircle(GeoPoint geoPt0, GeoPoint geoPt1, double lat,
                      double *lon0, double *lon1)
{
  /*
   *  Find the intersection of a great circle and a latitude circle, given
   *  two points on the great circle and the latitude.  Longitudes at which
   *  the great circle crosses are put into lons, which should have space for
   *  two floats.  Returns 1 if the two intersections exist.  Returns 0 if the
   *  great circle does not reach the indicated lat.
   *
   *  Plane for the great circle containing the two points is defined by
   *  (c.x, c.y, c.z) . (x, y, z) = 0
   *  c.x * cos(lat) * cos(lon) + c.y * cos(lat) * sin(lon) + c.z * sin(lat) = 0
   *  Put in lat, solve for lon.  Two solutions, go into lon0 and lon1.
   *  If lon0 != lon1, lon0 is between geoPt0 and geoPt1.
   */

  /* Geocentric Cartesian coordinates */
  CartPoint
    cpt0 = LatLonToCart(geoPt0),
    cpt1 = LatLonToCart(geoPt1),
    c;			/* Cross product of vectors to the two points */

  double lon;		/* Azimuth of the two points */
  double arg;		/* Computational thing */

  c.x = cpt0.y * cpt1.z - cpt0.z * cpt1.y,
  c.y = cpt0.z * cpt1.x - cpt0.x * cpt1.z,
  c.z = cpt0.x * cpt1.y - cpt0.y * cpt1.x;
  lon = atan2(c.y, c.x) * DEGPERRAD;

  if (   (arg = c.x * c.x + c.y * c.y) == 0.0
      || fabs(arg = -c.z * tan(lat * RADPERDEG) / sqrt(arg)) > 1.0)
    return 0;	/* Great circle does not reach lat */

  arg = acos(arg) * DEGPERRAD;
  *lon0 = GwchLon(lon + arg);
  *lon1 = GwchLon(lon - arg);
  if (LonBtwn(*lon1, geoPt0.abs, geoPt1.abs)) {
    /* We want lon0 to be between the GeoPoints */
    double tmpLon = *lon0;
    *lon0 = *lon1;
    *lon1 = tmpLon;
  }
  return 1;
}
