/*
 * tclGeoFile.c - use files with geographic information
 *
 * Copyright (C) 1999 Gordon D. Carrie
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 * 
 * Please address questions and bug reports to gcarri@mail.txcyber.com
 *
 ********************************************
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <tcl.h>
#include "tclGeoFile.h"
#include "geography.h"

/* Structure for reading GeoPoints from various streams */
struct GeoRdFile
{
  Tcl_Interp *interp;		/* Interpreter for error messages */
  union
    {
      struct			/* Standard stream */
        {
          Tcl_Channel chnl;	/* Tcl channel for io stream */
          FILE *file;		/* Standard io stream */
        } stdio;
      struct			/* XDR stream */
	{
          Tcl_Channel chnl;	/* Tcl channel for io stream for xdr stream */
	  FILE *file;		/* Standard io stream for xdr stream */
	  XDR xdrs;		/* XDR stream */
	} xdr;
    } stream;
  void (*free)(GeoRdFile geoRdFile);
  int (*getString)(GeoRdFile geoRdFile, char *string);
  int (*getInt)(GeoRdFile geoRdFile, int *i);
  int (*getGeoPoint)(GeoRdFile geoRdFile, GeoPoint *geoPoint);
};

/* Local functions */
static void freeStd(GeoRdFile geoRdFile);
static void freeXDR(GeoRdFile geoRdFile);
static int  getStringASCII(GeoRdFile geoRdFile, char *string);
static int  getIntASCII(GeoRdFile geoRdFile, int *i);
static int  getGeoPointASCII(GeoRdFile geoRdFile, GeoPoint *geoPoint);
static int  getStringXDR(GeoRdFile geoRdFile, char *string);
static int  getIntXDR(GeoRdFile geoRdFile, int *i);
static int  getGeoPointXDR(GeoRdFile geoRdFile, GeoPoint *geoPoint);

/* Memory management */
GeoRdFile
GeoRdFileAlloc()
{
  /* Allocate storage for a GeoRdFile */

  struct GeoRdFile *geoRdFile;		/* Hold return value */

  if ( !(geoRdFile = (struct GeoRdFile *)malloc(sizeof(struct GeoRdFile))))
    return NULL;
  memset(geoRdFile,0,sizeof(struct GeoRdFile));
  return geoRdFile;
}

void
GeoRdFileFree(GeoRdFile geoRdFile)
{
  /* Free all storage associated with a GeoRdFile */
  if (geoRdFile && geoRdFile->free)
    geoRdFile->free(geoRdFile);
}

/* Parameter setting routines */

int
GeoRdFileASCII(Tcl_Interp *interp, GeoRdFile geoRdFile, char *name)
{
  /*
   * Open a GeoRdFile containing an ASCII stream.  If name starts with '|',
   * opens a pipe.
   */

  char **argv;			/* Argument list for pipe command */
  int
    argc,			/* Argument count for pipe command */
    fd;				/* File descriptor */

  geoRdFile->interp = interp;

  /* Open the input channel */
  if (*name == '|')
    {
      /* Name is a command */
      if (Tcl_SplitList(interp, name + 1, &argc, &argv) == TCL_ERROR
          || !(geoRdFile->stream.stdio.chnl
               = Tcl_OpenCommandChannel(interp, argc, argv, TCL_STDOUT)))
          return TCL_ERROR;
      free(argv);
    }
  else
    {
      /* Name is a file name */
      if (!(geoRdFile->stream.stdio.chnl
            = Tcl_OpenFileChannel(interp, name, "r", 0)))
        return TCL_ERROR;
    }

  /* Get FILE for channel */
  if (Tcl_GetChannelHandle(geoRdFile->stream.stdio.chnl, TCL_READABLE,
                           (ClientData *)&fd) == TCL_ERROR
      || !(geoRdFile->stream.stdio.file = fdopen(fd, "r")))
    {
      Tcl_SetResult(interp, "Could not convert channel to file", TCL_STATIC);
      return TCL_ERROR;
    }

  geoRdFile->free = freeStd;
  geoRdFile->getString = getStringASCII;
  geoRdFile->getInt = getIntASCII;
  geoRdFile->getGeoPoint = getGeoPointASCII;
  return TCL_OK;
}

int
GeoRdFileXDR(Tcl_Interp *interp, GeoRdFile geoRdFile, char *name)
{
  /* Open a GeoRdFile containing an XDR stream */

  char **argv;			/* Argument list for pipe command */
  int
    argc,			/* Argument count for pipe command */
    fd;				/* File descriptor */

  geoRdFile->interp = interp;

  /* Open the input channel */
  if (*name == '|')
    {
      /* Name is a command */
      if (Tcl_SplitList(interp, name + 1, &argc, &argv) == TCL_ERROR
          || !(geoRdFile->stream.xdr.chnl
               = Tcl_OpenCommandChannel(interp, argc, argv, TCL_STDOUT)))
          return TCL_ERROR;
      free(argv);
    }
  else
    {
      /* Name is a file name */
      if (!(geoRdFile->stream.xdr.chnl
            = Tcl_OpenFileChannel(interp, name, "r", 0)))
        return TCL_ERROR;
    }

  /* Get FILE for channel */
  if (Tcl_GetChannelHandle(geoRdFile->stream.xdr.chnl, TCL_READABLE,
                           (ClientData *)&fd) == TCL_ERROR
      || !(geoRdFile->stream.xdr.file = fdopen(fd, "r")))
    {
      Tcl_SetResult(interp, "Could not convert channel to file", TCL_STATIC);
      return TCL_ERROR;
    }
  xdrstdio_create(&geoRdFile->stream.xdr.xdrs, geoRdFile->stream.xdr.file,
		  XDR_DECODE);
  geoRdFile->free = freeXDR;
  geoRdFile->getString = getStringXDR;
  geoRdFile->getInt = getIntXDR;
  geoRdFile->getGeoPoint = getGeoPointXDR;
  return TCL_OK;
}

/* Reading routines */
int
GeoRdFileString(GeoRdFile geoRdFile, char *string)
{
  return geoRdFile->getString(geoRdFile, string);
}

int GeoRdFileInt(GeoRdFile geoRdFile, int *i)
{
  return geoRdFile->getInt(geoRdFile, i);
}

int
GeoRdFileGeoPoint(GeoRdFile geoRdFile, GeoPoint *geoPoint)
{
  return geoRdFile->getGeoPoint(geoRdFile, geoPoint);
}

/* End of global function definitions */

void
freeStd(GeoRdFile geoRdFile)
{
  Tcl_Close(geoRdFile->interp, geoRdFile->stream.stdio.chnl);
  free(geoRdFile);
  geoRdFile = NULL;
}

void
freeXDR(GeoRdFile geoRdFile)
{
  xdr_destroy(&geoRdFile->stream.xdr.xdrs);
  Tcl_Close(geoRdFile->interp, geoRdFile->stream.xdr.chnl);
  free(geoRdFile);
  geoRdFile = NULL;
}

int
getStringASCII(GeoRdFile geoRdFile, char *string)
{
  /*
   * Read a character string from an ASCII stream into string.
   * First reads number of characters, as decimal integer.
   * If string is NULL, allocates that many bytes + 1 for it,
   * else if string is not NULL, reallocates that many bytes + 1.
   * string must be freed by the user
   */

  int nChar;			/* Number of bytes in string */
  char fmt[14];			/* Format specifier for reading string */

  if (fscanf(geoRdFile->stream.stdio.file, "%d", &nChar) != 1)
    return 0;
  if ( (!string && !(string = (char *)malloc(nChar + 1)))
      || !(string = (char *)realloc(string, nChar + 1)))
    {
      return 0;
    }
  sprintf(fmt, " %%%dc", nChar);
  if (fscanf(geoRdFile->stream.stdio.file, fmt, string) != 1)
    return 0;
  string[nChar] = '\0';
  return 1;
}

int
getIntASCII(GeoRdFile geoRdFile, int *i)
{
  return (fscanf(geoRdFile->stream.stdio.file, "%d", i) == 1);
}

int
getGeoPointASCII(GeoRdFile geoRdFile, GeoPoint *geoPoint)
{
  /* Read latitude and longitude into geoPoint.  May be surrounded by () */
  
  return ( fscanf(geoRdFile->stream.stdio.file, "%f%f",
		  &geoPoint->ord, &geoPoint->abs) == 2
	  || fscanf(geoRdFile->stream.stdio.file, "( %f  %f )",
		    &geoPoint->ord, &geoPoint->abs) == 2);
}

int
getStringXDR(GeoRdFile geoRdFile, char *string)
{
  /*
   * Get a string from an xdr stream
   * If string is NULL, it is allocated, otherwise is reallocated.
   * Must be freed by user.
   */

  char *tmpStr = NULL;		/* Hold temporary string */
  int nChar;			/* Number of non-null char's in input string */

  if ( !xdr_string(&geoRdFile->stream.xdr.xdrs, &tmpStr, INT_MAX))
    return 0;
  nChar = strlen(tmpStr);
  if ( (!string && !(string = (char *)malloc(nChar + 1)))
      || !(string = (char *)realloc(string, nChar + 1)))
    {
      return 0;
    }
  strncpy(string, tmpStr, nChar);
  string[nChar] = '\0';
  free(tmpStr);
  return 1;
}

int
getIntXDR(GeoRdFile geoRdFile, int *i)
{
  return xdr_int(&geoRdFile->stream.xdr.xdrs, i);
}

int
getGeoPointXDR(GeoRdFile geoRdFile, GeoPoint *geoPoint)
{
  float ord, abs;
  if ( !(   xdr_float(&geoRdFile->stream.xdr.xdrs, &ord)
         && xdr_float(&geoRdFile->stream.xdr.xdrs, &abs))) 
    return 0;
  geoPoint->ord = ord;
  geoPoint->abs = abs;
  return 1;
}

