/*
 * tclGeoLines.c - maintain a database of latLonArrays
 *
 * Copyright (C) 1999 Gordon D. Carrie
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 * 
 * Please address questions and bug reports to gcarri@mail.txcyber.com
 *
 **********************************************************************
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <limits.h>
#include <float.h>
#include <tcl.h>
#include "geography.h"
#include "geoLines.h"
#include "tclGeoLines.h"
#include "tclGeoFile.h"

#define GEOLINES "geolinearray"

static Tcl_HashTable lineCmds;	/* Table of commands */
static int loaded;		/* Tell if package already loaded */
static char result[200];	/* Place to format result */

/* Table of GeoLineArrays */
/* Keys are array names like "continents" or "counties" */
/* Values are pointers to GeoLineArrays */
static Tcl_HashTable latLonArrays;

/* Local functions */
static Tcl_CmdProc geoLineCmd;
static Tcl_CmdProc newLatLonArray;	/* Create an array of lines */
static Tcl_CmdProc geoLinesFmList;	/* Create array of lines from list */
static Tcl_CmdProc rmLatLonArray;	/* Delete an array of lines */
static Tcl_CmdProc listLatLonArrays;	/* Return a list of LatLonArray names*/
static Tcl_CmdProc latLonArrayLimits;	/* Return array limits */
static Tcl_CmdProc cross;		/* Find intersection of two segments*/
static Tcl_CmdProc geoPtInLineArray;	/* Tell if point bound by line array */
static GeoLineArray geoLinesLoad(Tcl_Interp *interp, char *cmd,
         GeoRdFile geoRdFile);		/* Load lines from a geoRdFile */

/******* End of declarations ******/

int
Tclgeolines_Init(Tcl_Interp *interp)
{
  /*
   * LatLonsInit - create the table to map geoLine names to geoLine structs
   *   and create the commands to access it
   * Input: interp - tcl interpreter
   * Return value:
   *   TCL_OK or TCL_ERROR
   */

  /* Struct to map command name to command function */
  struct NamedCommand
  {
    char	*name;
    Tcl_CmdProc	*function;
  };

  /* GeoLine commands and their functions */
  struct NamedCommand geoLineCmdList[] =
    {
      {"fmxdr",		newLatLonArray	},
      {"fmascii",	newLatLonArray},
      {"fmlist",	geoLinesFmList},
      {"rm",		rmLatLonArray	},
      {"remove",	rmLatLonArray	},
      {"list",		listLatLonArrays},
      {"limits",	latLonArrayLimits},
      {"cross",		cross},
      {"containpt",	geoPtInLineArray},
      {NULL,		NULL}
    };

  struct NamedCommand *command;	/* Loop pointer */
  int np;

  if (loaded)
    return TCL_OK;
  loaded = 1;

  /* Create the geolines command and subcommands */
  Tcl_CreateCommand(interp, GEOLINES, geoLineCmd, NULL,
		    (Tcl_CmdDeleteProc *)NULL);
  Tcl_InitHashTable(&lineCmds, TCL_STRING_KEYS);
  for (command = geoLineCmdList; command->name; command++)
    Tcl_SetHashValue(Tcl_CreateHashEntry(&lineCmds, command->name, &np),
		     (ClientData)command->function);
  Tcl_InitHashTable(&latLonArrays, TCL_STRING_KEYS);
  Tcl_PkgProvide(interp, "tclgeolines", "1.2");
  return TCL_OK;
}

int
geoLineCmd(ClientData clientData, Tcl_Interp *interp, int argc,
	       char **argv)
{
  /*
   * geoLineCmd - Find and execute geoline subcommand
   * Usage:
   *   geolines subcommand ?option arg? ?option arg? ... 
   * Input:
   *   clientData - not used
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Return value:
   *   Result of executed command, or TCL_ERROR.
   * Side effects:
   *   Calls the function corresponding to command named on the command line
   *   Appends list of geolines commands to interp->result on bad call
   */
   
  int result;
  Tcl_HashEntry *entry;
  Tcl_CmdProc *command;
  Tcl_HashSearch search;

  if ((argc == 1) || !(entry = Tcl_FindHashEntry(&lineCmds, argv[1])))
    {
      /* Provide list of available commands */
      Tcl_AppendResult(interp, argv[0], " commands are: ", (char *)NULL);
      for (entry = Tcl_FirstHashEntry(&lineCmds, &search);
	   entry != NULL;
	   entry = Tcl_NextHashEntry(&search))
	Tcl_AppendResult(interp, Tcl_GetHashKey(&lineCmds, entry), " ",
			 (char*)NULL);
      Tcl_AppendResult(interp, "\n", (char *)NULL);

      result = TCL_ERROR;
    }
  else
    {
      command = (Tcl_CmdProc *)Tcl_GetHashValue(entry);
      result = (*command)(NULL, interp, argc, argv);
    }
  return result;
}

GeoLineArray
GetLatLonArray(char *latLonArrayName)
{
  /* Return GeoLineArray named lineArrayName */
  Tcl_HashEntry *entry;
  if ( !(entry = Tcl_FindHashEntry(&latLonArrays, latLonArrayName)))
    return NULL;
  return Tcl_GetHashValue(entry);
}

int
GeoLatLonUseArray(char *latLonArrayName)
{
  /*
   * Increment userCnt for LatLonArray identified in
   * lineArray table by latLonArrayName.
   * You can later decrement userCnt with LatLonFreeArray.
   * Must have userCnt==0 before deleting array
   */

  Tcl_HashEntry *entry;

  if ( !(entry = Tcl_FindHashEntry(&latLonArrays, latLonArrayName)))
    return TCL_ERROR;
  GeoLineArrayAddUser(Tcl_GetHashValue(entry));
  return TCL_OK;
}

void
GeoLatLonFreeArray(char *latLonArrayName)
{
  /*
   * GeoLatLonFreeArray - decrement userCnt for latLonArray identified
   * in lineArray table by latLonArrayName.
   * Must have userCnt==0 before deleting array
   */
   
  Tcl_HashEntry *entry;

  if ( !(entry = Tcl_FindHashEntry(&latLonArrays, latLonArrayName)))
    return;
  GeoLineArrayDropUser(Tcl_GetHashValue(entry));
  return;
}

int
newLatLonArray(ClientData clientData, Tcl_Interp *interp, int argc,char *argv[])
{
  /*
   * latLonArray - read and store an array of LatLons as lat's and lon's
   * Usage: geolines fmxdr|fmascii latLonArrayName fileName 
   * Input:
   *   clientData - not used
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Return value:
   *   TCL_OK or TCL_ERROR
   * Side effects:
   *   Creates an array of geolines and adds its name to the latLonArrays table.
   *   Appends error message to interp->result on failure
   */
  
  char
    *latLonArrayName,			/* Table key */
    *fileName,				/* Name of input stream */
    *fileType;				/* Type of input stream (ascii, xdr) */
  Tcl_HashEntry *entry;			/* Entry for new latLonArray */
  int newPtr;				/* Indicate if latLonArray exists*/
  GeoRdFile geoRdFile = NULL;		/* Input stream */
  unsigned nLinesMax = 10;		/* Initial # of lines in latLonArray */
  GeoLineArray latLonArray = NULL;	/* Storage for new latLonArray */

  if (argc != 4) {
    Tcl_AppendResult(interp, "Usage: \"", argv[0], " ", argv[1],
      " latLonArrayName fileName\"\n", (char *)NULL);
    return TCL_ERROR;
  }

  fileType = argv[1];
  latLonArrayName = argv[2];
  fileName = argv[3];

  /* Create new table entry */
  entry = Tcl_CreateHashEntry(&latLonArrays, latLonArrayName, &newPtr);
  if (newPtr == 0) {
    Tcl_AppendResult(interp, latLonArrayName, " already exists.\n",
		       (char *)NULL);
    return TCL_ERROR;
  }

  /* Allocate storage for new array */
  if ( !(latLonArray = NewGeoLineArray(nLinesMax)) ) {
    Tcl_AppendResult(interp, "Allocation failed for ", latLonArrayName, "\n",
      (char *)NULL);
    Tcl_DeleteHashEntry(entry);
    return TCL_ERROR;
  }

  /* Read lines from file or stream */
  if (   (strcmp(fileType, "fmxdr") == 0
          && !(   (geoRdFile = GeoRdFileAlloc())
               && GeoRdFileXDR(interp, geoRdFile, fileName) == TCL_OK))
      || (strcmp(fileType, "fmascii") == 0
          && !(   (geoRdFile = GeoRdFileAlloc())
	       && GeoRdFileASCII(interp, geoRdFile, fileName) == TCL_OK))) {
     
    goto error;
   
  } else {

    /* Read in the lines */
    if ( !(latLonArray = geoLinesLoad(interp, argv[0], geoRdFile)) ) {
      goto error;
    } else {
      Tcl_SetHashValue(entry, latLonArray);
      Tcl_SetResult(interp, latLonArrayName, TCL_VOLATILE);
    }
    GeoRdFileFree(geoRdFile);

  }
  return TCL_OK;
  
 error:
  Tcl_AppendResult(interp, "Could not load lines from ", argv[3], "\n",
		   (char *)NULL);
  FreeGeoLineArray(latLonArray);
  free(latLonArray);
  Tcl_DeleteHashEntry(entry);
  GeoRdFileFree(geoRdFile);
  return TCL_ERROR;
}

int
geoLinesFmList(ClientData clientData, Tcl_Interp *interp,
		   int argc, char *argv[])
{
  /*
   * geoLinesFmList - get a geoLineArray from a list of lists on command line.
   * Input:
   *   clientData - not used
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Return value:
   *   TCL_OK or TCL_ERROR
   * Side effects:
   *   Creates an array of geolines and adds its name to the latLonArrays table.
   *   Appends error message to interp->result on failure
   *
   * Either argv[3] is a list of GeoPoints in a single GeoLine
   *    e.g. {{lat0 lon0} {lat1 lon1} {lat2 lon2} {lat3 lon3}}
   *
   * or argv[3] is a list of lists containing geolines.
   *    e.g { { {lat00 lon00} {lat01 lon01} {lat02 lon02} {lat03 lon03} }
   *          { {lat10 lon10} {lat11 lon11} {lat12 lon12} } }
   *    has two lines, the first with four points, the second with three.
   */

  char
    *latLonArrayName,			/* Table key */
    *list;				/* List of lines.  Each a list of pts*/
  Tcl_HashEntry *entry;			/* Entry for new latLonArray */
  int newPtr;				/* Indicate if latLonArray exists*/
  GeoLineArray latLonArray = NULL;	/* Storage for new latLonArray */
  char
    *errmsg,				/* Error message  */
    **elems,				/* Array of lines or points */
    **lines,				/* Array of lines (as text strings) */
    **points;				/* Array of lat-lons */
  int 
    n,					/* Loop index */
    nLines,				/* Number of lines */
    npts,				/* Number of points in a line*/
    nl;					/* Loop index */
  GeoLine geoLine = NULL;		/* Hold current geoLine */
  GeoPoint geoPoint;			/* Scanned point */

  /************************** End of declarations **************************/

  if (argc < 4) {
    Tcl_AppendResult(interp, "Usage: \"", argv[0], " ", argv[1],
      " latLonArrayName list \"\n", (char *)NULL);
    return TCL_ERROR;
  }

  latLonArrayName = argv[2];
  list = argv[3];

  /* Create new table entry */
  entry = Tcl_CreateHashEntry(&latLonArrays, latLonArrayName, &newPtr);
  if (newPtr == 0) {
    Tcl_AppendResult(interp, latLonArrayName, " already exists.\n",
		       (char *)NULL);
    return TCL_ERROR;
  }

  /* Allocate geoLine */
  if ( !(geoLine = NewGeoLine(0)) ) {
    errmsg = "Could not allocate geoLine";
    goto error;
  }

  /* Get the list of GeoLines or GeoPoints */
  if (Tcl_SplitList(interp, list, &n, &elems) == TCL_ERROR || n == 0) {
    errmsg = "Could not split list";
    goto error;
  }

  /* Read list of GeoPoints or list of GeoLines */
  if (sscanf(elems[0], "%f %f", &geoPoint.ord, &geoPoint.abs) == 2) {

    /* First list element is a GeoPoint => list is a singe GeoLine */
    if ( (npts = n) < 2 ) {
      errmsg = "Line cannot have only one point";
      goto error;
    }

    points = elems;
    if ( !(latLonArray = NewGeoLineArray(1)) ) {
      errmsg = "Could not allocate latLonArray";
      goto error;
    }
    for (n = 0; n < npts; n++) {
      if (sscanf(points[n], "%f %f", &geoPoint.ord, &geoPoint.abs) != 2) {
        Tcl_AppendResult(interp, argv[0], ": Could not read point ",
	    points[n], "\n", (char *)NULL);
        errmsg = "Unable to read list of points";
	goto error;
      }
      if ( !AppendGeoPoint(geoPoint, geoLine) ) {
        errmsg = "Unable to append point";
	goto error;
      }
    }
    if ( !AppendGeoLine(geoLine, latLonArray) ) {
      errmsg = "Could not append line to array";
      goto error;
    }
    free(points);

  } else {

    /* First list element not a GeoPoint => list is list of GeoLines */
    nLines = n;
    lines = elems;
    if ( !(latLonArray = NewGeoLineArray(nLines)) ) {
      errmsg = "Could not allocate latLonArray";
      goto error;
    }

    /* Read each GeoLine from the list of GeoLines */
    /* (each of which is a list of geoPoints)	   */
    for (nl = 0; nl < nLines; nl++) {
      if (Tcl_SplitList(interp, lines[nl], &npts, &points) == TCL_ERROR) {
        errmsg = "Could not split list of points";
        goto error;
      }
      if (npts < 2) {
        errmsg = "Line cannot have only one point";
        goto error;
      }
      for (n = 0; n < npts; n++) {
        if (sscanf(points[n], "%f %f", &geoPoint.ord, &geoPoint.abs) != 2) {
          Tcl_AppendResult(interp, argv[0], ": Could not read point ",
	    points[n], "\n", (char *)NULL);
          errmsg = "Unable to read list of points";
  	goto error;
        }
        if ( !AppendGeoPoint(geoPoint, geoLine) ) {
          errmsg = "Unable to append point";
  	goto error;
        }
      }
      if ( !AppendGeoLine(geoLine, latLonArray) ) {
        errmsg = "Could not append line to array";
        goto error;
      }
      ClearGeoLine(geoLine);	/* Reset geoLine */
      free(points);
    }
    free(lines);
  }

  FreeGeoLine(geoLine);
  Tcl_SetHashValue(entry, latLonArray);
  Tcl_SetResult(interp, latLonArrayName, TCL_VOLATILE);
  return TCL_OK;

  /* Error return */
  error:
    Tcl_AppendResult(interp, argv[0], ": ", errmsg, "\n", (char *)NULL);
    Tcl_DeleteHashEntry(entry);
    FreeGeoLineArray(latLonArray);
    return TCL_ERROR;
}

int
rmLatLonArray(ClientData clientData, Tcl_Interp *interp, int argc,
		  char *argv[])
{
  /*
   * rmLatLonArray - remove a LatLonArray
   * Usage: geolines rm latLonArrayName
   * Input:
   *   clientData - not used
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Return value:
   *   TCL_OK or TCL_ERROR
   * Side effects:
   *   Frees a latLonArray and removes it's name from the latLonArray table.
   *   Appends error message to interp->result on failure
   */

  Tcl_HashEntry *entry;		/* Entry for geoLine */
  GeoLineArray latLonArray;	/* Storage for latLonArray */

  if (argc != 3) {
    Tcl_AppendResult(interp,
      "Usage: \"", argv[0], " ", argv[1], " latLonArrayName\"\n", (char *)NULL);
    return TCL_ERROR;
  }
  if ((entry = Tcl_FindHashEntry(&latLonArrays, argv[2])) != NULL) {
    latLonArray = Tcl_GetHashValue(entry);
    if (GeoLineArrayUserCnt(latLonArray)) {
      Tcl_AppendResult(interp, argv[2], " in use.\n", (char *)NULL);
      return TCL_ERROR;
    }
    FreeGeoLineArray(latLonArray);
    Tcl_DeleteHashEntry(entry);
  }
  return TCL_OK;
}

int
listLatLonArrays(ClientData clientData, Tcl_Interp *interp,	int argc, 
                     char *argv[])
{
  /*
   * listLatLonArrays - put a list of latLonArrays into the result
   * Usage: geolines fmxdr|fmascii latLonArrayName fileName
   * Input:
   *   clientData - not used
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Return value:
   *   TCL_OK or TCL_ERROR
   */

  Tcl_HashEntry *entry;
  Tcl_HashSearch search;

  for (entry = Tcl_FirstHashEntry(&latLonArrays, &search);
       entry != NULL;
       entry = Tcl_NextHashEntry(&search))
    {
      Tcl_AppendElement(interp, Tcl_GetHashKey(&latLonArrays, entry));
    }

  return TCL_OK;
}

int
latLonArrayLimits(ClientData clientData, Tcl_Interp *interp, int argc, 
                     char *argv[])
{
  /*
   * latLonArrayLimits - set result to lat lon extremes of a line array
   *   Printed as:  ordMax absMax ordMin absMin
   *   i.e. clockwise from north.
   * Usage: geolines limits lineArrayName
   * Input:
   *   clientData - not used
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Return value:
   *   TCL_OK or TCL_ERROR
   * Side effects:
   *   Sets result to 4 lat-lon values or appends an error message.
   */
   
  Tcl_HashEntry *entry;		/* Entry for geoLine */
  GeoLineArray latLonArray;	/* Storage for latLonArray */

  if (argc != 3) {
    Tcl_AppendResult(interp, "Usage: \"", argv[0], " ", argv[1],
      " latLonArrayName\"\n", (char *)NULL);
    return TCL_ERROR;
  }
  if ( !(entry = Tcl_FindHashEntry(&latLonArrays, argv[2]))) {
    Tcl_AppendResult(interp, "No line array named ", argv[2], "\n",
      (char *)NULL);
    return TCL_ERROR;
  } else {
    latLonArray = Tcl_GetHashValue(entry);
    sprintf(result, "%8.3f %8.3f %8.3f %8.3f",
      GeoLineArrayOrdMax(latLonArray), GeoLineArrayAbsMax(latLonArray),
      GeoLineArrayOrdMin(latLonArray), GeoLineArrayAbsMin(latLonArray));
    Tcl_SetResult(interp, result, TCL_VOLATILE);
  }
  return TCL_OK;  
}

int
cross(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /*
   * cross - find intersection of two lines given two points on each
   *   segment.
   *
   * Usage:
   *   geolines cross {line1_lat1 line1_lon1} {line1_lat2 line1_lon2} \
   *     {line2_lat1 line2_lon1} {line2_lat2 line2_lon2}
   *
   * {line1_lat1 line1_lon1} {line1_lat2 line1_lon2} are two points on 1st line
   * {line2_lat1 line2_lon1} {line2_lat2 line2_lon2} are two points on 2nd line
   *
   * Input:
   *   clientData - not used
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Return value:
   *   TCL_OK or TCL_ERROR
   * Side effects:
   *   On success, sets result to two element list containing longitude and
   *   latitude of intersection of the two lines.  Note:  Great circles have
   *   two intersections.  Returns the one that is "closer."
   *   On failure, sets result to an error message.
   *
   */

  GeoPoint
    ln1Pt1, ln1Pt2,	/* Points on 1st line */
    ln2Pt1, ln2Pt2,	/* Points on 2nd line */
    geoPts[4],		/* Array to use when reading */
    geoPt;		/* Hold return value */

  int a, p;		/* Counters */

  /* Parse the command line */
  if (argc < 6) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
     " {ln1_lat1 ln1_lon1} {ln1_lat2 ln1_lon2}"
     "  {ln2_lat1 ln2_lon1} {ln2_lat2 ln2_lon2}", NULL);
    return TCL_ERROR;
  }
  for (a = 2, p = 0; a < argc; a++) {
    if (sscanf(argv[a], "%f %f", &geoPts[p].ord, &geoPts[p].abs) == 2) {
      p++;	/* Have read next GeoPoint */
    } else {
      Tcl_AppendResult(interp, argv[0], " ", argv[1], " has bad arg: ",
        argv[a], (char *)NULL);
      return TCL_ERROR;
    }
  }

  ln1Pt1 = geoPts[0];
  ln1Pt2 = geoPts[1];
  ln2Pt1 = geoPts[2];
  ln2Pt2 = geoPts[3];

  geoPt = GCircleX(ln1Pt1, ln1Pt2, ln2Pt1, ln2Pt2);

  if (geoPt.ord > 90.0) {
    Tcl_AppendResult(interp, argv[0], " ", argv[1], " got undefined "
      "intersection", (char *)NULL);
    return TCL_ERROR;
  } else {
    sprintf(result, "%g %g", geoPt.ord, geoPt.abs);
    Tcl_SetResult(interp, result, TCL_VOLATILE);
  }

  return TCL_OK;
}

int
geoPtInLineArray(ClientData clientData, Tcl_Interp *interp, int argc,
                   char *argv[])
{
  /* Tell if point bound by line array
   * Usage:
   *   geolines containpt latLonArrayName geoPt
   * Return value:
   *   TCL_OK or TCL_ERROR
   * Side effects:
   *   Sets result to "1" if any lineArray in latLonArray contains geoPt,
   *   or sets result to "0" if no lineArray in latLonArray contains geoPt.
   */

  Tcl_HashEntry *entry;		/* Entry for geoLine */
  GeoLineArray latLonArray;	/* Storage for latLonArray */
  GeoPoint geoPt;		/* Point to determine */

  /* Check command line */
  if (argc != 4) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " containpt latLonArrayName "
      "geoPt", NULL);
    return TCL_ERROR;
  }

  /* Get lineArray */
  if ( !(entry = Tcl_FindHashEntry(&latLonArrays, argv[2]))) {
    Tcl_AppendResult(interp, "No line array named ", argv[2], "\n",
      (char *)NULL);
    return TCL_ERROR;
  } else {
    latLonArray = Tcl_GetHashValue(entry);
    GeoLineArrayOrdMin(latLonArray), GeoLineArrayAbsMin(latLonArray);
  }

  /* Get point */
  if ( sscanf(argv[3], "%f %f", &geoPt.ord, &geoPt.abs) != 2) {
    Tcl_AppendResult(interp, argv[0], " could not read ", argv[3], " as a "
      " lat-lon", NULL);
    return TCL_ERROR;
  }

  /* Find out if point is contained in any line in lineArray */
  if ( GeoLineArrayContainGeoPoint(geoPt, latLonArray) ) {
    Tcl_SetResult(interp, "1", TCL_STATIC);
  } else {
    Tcl_SetResult(interp, "0", TCL_STATIC);
  }

  return TCL_OK;  

}

GeoLineArray
geoLinesLoad(Tcl_Interp *interp, char *cmd, GeoRdFile geoRdFile)
{
  /*
   * geoLinesLoad - get a geoLineArray from a geoRdFile
   * Input:
   *   interp    - tcl interpreter for error messages
   *   cmd       - name of calling command, to use in error messages
   *   geoRdFile - Pointer to a GeoRdFile struct
   *     Input stream is assumed to be in the following format:
   *	 String descriptor
   *     Then for each line:
   *       Number of points for the line
   *       Latitude longitude pairs for the line
   * Return: array of Line structs, or nullArray if there is a failure
   */

  char
    *descr = NULL,		/* Descriptor */
    *errmsg;			/* Error message  */
  int
    n,				/* Loop index */
    npts;			/* Number of points to read in */
  GeoLineArray lineArray;	/* Hold return value */
  GeoLine line;			/* Hold input GeoLine */
  GeoPoint pt;			/* Hold input GeoPoint */
  unsigned
    nptsMax = 100,		/* Initial number of points in line */
    nLinesMax = 100;		/* Initial number of lines in array */

  /************************** End of declarations **************************/

  /* Initialize line and lineArray */
  if ( !(lineArray = NewGeoLineArray(nLinesMax)) )
    return NULL;
  if ( !(line = NewGeoLine(nptsMax)) ) {
    FreeGeoLineArray(lineArray);
    return NULL;
  }

  /* Get the descriptor and throw it away */
  if ( !GeoRdFileString(geoRdFile, descr) ) {
    errmsg = "Read descriptor failed";
    goto error;
  }

  /* Read in the lines */
  while (GeoRdFileInt(geoRdFile, &npts)) {	/* Read npts for next line */

    /* Read points */
    for (n = 0; n < npts; n++) {
      if ( !GeoRdFileGeoPoint(geoRdFile, &pt) ) {
        errmsg = "Read GeoPoint failed";
        goto error;
      }
      if ( !AppendGeoPoint(pt, line) ) {
        errmsg = "Append GeoPoint failed";
        goto error;
      }
    }

    /* Add line to array */
    if ( !AppendGeoLine(line, lineArray) ) {
      errmsg = "Could not add new line to lineArray";
      goto error;
    }
    ClearGeoLine(line);		/* Reset input line */
  }

  /* Eliminate wasted space in lineArray */
  if ( !GrowGeoLineArray(lineArray, GeoLineArrayNLines(lineArray)) ) {
    errmsg = "Could not adjust size of lineArray\n";
    goto error;
  }

  FreeGeoLine(line);
  return lineArray;

  error:
    Tcl_AppendResult(interp, cmd, ": ", errmsg, "\n", (char *)NULL);
    FreeGeoLineArray(lineArray);
    FreeGeoLine(line);
    return NULL;
}
